#include "main-c.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/*#include <unistd.h>*/
#include <sys/types.h>
#include <asm/byteorder.h>

#include "sim_gtk.h"

#ifndef VERSION
#define VERSION "0.8.7"
#endif

#define PLUG_IN_VERSION VERSION" - 1998-10-08"

#define DEBUGLEVEL	0

#if (DEBUGLEVEL > 0)
#	define VERBOSE		
#endif

#if (DEBUGLEVEL > 1)
#	define DEBUG		
#endif

/*
 *  This is the IFF-ILBM Gimp plugin by Johnny Teveen,
 *  eMail: <j.tevessen@line.org>
 *
 *  Short:
 *    IFF (InterchangeFileFormat) is a universal data
 *    exchange format developed by Electronic Arts.
 *    It has shipped with the legendary Amiga computers
 *    for years. IFF-ILBM (InterLeavedBitMap) is a
 *    subtype for storing bitmapped images in a non-
 *    chunky, line oriented format.  Great for pictures
 *    upto 256 colors, quite tricky for 24bit ones.
 *      IFF is very extensible; programs shall skip
 *    unknown chunks (blocks of information).  Different
 *    types of data can be stored in the same file
 *    using a recursive or stringent structure.  This
 *    feature is rarely used.  Compression (using a
 *    runlength algorithm) is bad.
 *      Being born into the Amiga custom chipset, IFF-
 *    ILBM supports a CAMG (CommodoreAMiGa) chunk
 *    containing Amiga native flags for the screenmode
 *    (resolution and decoding algorithm) to use.
 *    This includes HAM (Hold-And-Modify, a choice-
 *    reduced 4096..16M mode), EHB (Extra-HalfBrite,
 *    supports twice as many colors, but the upper
 *    half is a mirror of the lower half in 50%
 *    intensity), and asymmetric LoRes, Interlaced,
 *    HiRes, and SuperHiRes modes.  This allowes
 *    geometries of 1:1, 1:2, 1:4, and 2:1.  Two
 *    aspect fields contain more exact informations,
 *    but they're often not fulfilled correctly.
 *      If CAMG is missing, for example, one can
 *    only guess for HAM.
 *
 *  Already there:
 *
 *  o Support for uncompressed and ByteRun1 compressed
 *    bitmaps
 *  o Loading/Saving indexed pictures 1..256 colors
 *  o Loading/Saving transparency (alpha channel) in
 *    1bit-Mode (mask). Loading transparentColor mask.
 *  o Loading/Saving 24bit RGB pictures with 1bit
 *    alpha channel (25 planes in file)
 *  o Promoting indexed pictures to grayscale if possible
 *  o Auto-converting grayscale pictures to 256 color
 *    indexed mode on saving (convert image to indexed
 *    before saving to reduce amount of "colors")
 *  o Loading of HAM6 (as RGB) and EHB6 (as indexed64)
 *  o HAM6 transparency
 *
 *
 *  Missing:
 *
 *  o bmhd.masking == mskLasso (evil without a real blitter,
 *    need to do XOR throughout lines)
 *  o On indexed images, transparent color is ignored
 *    while loading
 *  o mskHasTransparentColor and HAM: Check whether it is okay
 *    to affect only the pixels directly drawn with that pen,
 *    or also add HAMified ones with the same color
 *  o Support for compression mode 2 (which has never really
 *    been important---it compresses column-by-column and
 *    compresses slightly better than ByteRun1)
 *  o Save options requester
 *  o Save HAM/EHB?
 *  o Check for 15/16 bit HighColor ILBMs
 *  o Improve IFF parsing
 *  o Add support for multiple images via catalogs
 *  o Somehow integrate ILBM-TINY thumbnails
 *  o Try to guess HAM if no CAMG chunk is given
 *  o Load option to allow scaling of LoRes/Interlaced-Images
 *  o Add HAM8 support
 *  o Add HAM3-5, HAM7, EHB2-5 support (yes, they're out there!)
 *  o Add aspectX/Y support (currently 1/1)
 *  o What about *real* alpha channels? 32planes, mask set?
 *  o Maybe integrate DRNG cycling chains as gradients
 *  o Could also be used to read ILBM palettes
 *
 *
 *  Used code fragments:
 *
 *  o ByteRun1 encode/decoder:
 *    Parts lent from packer.c/unpacker.c by Jerry Morrison
 *    and Steve Shaw, Electronic Arts. This code is placed
 *    in the Public Domain, dated 1985-11-15
 *  o GIMP plugin skeleton:
 *    Code of xpm.c (GPL)1995 Spencer Kimball and Peter Mattis
 *    and png.c (GPL)1998 Michael Sweet and Daniel Skarda
 *    used as a template
 *
 *  This code went through GNU indent(1).
 *  Please report bugs and suggestions.
 */


/**** Prototypes ****/

int progress_bar(int,int);

void ilbm_set_size(void *thiz, int w, int h, int exRGB, int isG, int alpha);
void ilbm_set_cmap(void *thiz, guchar *cmap, int ncols);
void ilbm_put(void *thiz, guchar *buf, int line);

/**** Plugin data ****/


typedef struct {
  gdouble threshold;
  gint32 compress;
  gint32 save_ham;
  gint32 set_backgnd;
} ILBMSaveVals;

typedef struct {
  gint run;
} ILBMSaveInterface;


/**** Types ****/

typedef guchar grayval;
typedef guchar palidx;          /* 8bit!! indexed files are built of it */

/**** Misc ****/

enum ColorLimit {
  maxIndexedColors = 256, maxGrayshades = maxIndexedColors
};

enum ByteWidth {
  byteppGray = 1, byteppGrayA = byteppGray + byteppGray,
  byteppRGB = byteppGray * 3, byteppRGBA = byteppRGB + byteppGray
};
enum BitWidth {
  bitppGray = byteppGray * 8 /* sizeof (grayval) */ , bitppGrayA = bitppGray + bitppGray,
  bitppRGB = byteppRGB * 3, bitppRGBA = bitppRGB + bitppGray
};

#define BYTEPL(onePlanePixelWidth) ((onePlanePixelWidth+7)>>3)

enum {
  transparent = 0x00, opaque = 0xFF
};

typedef struct {
  u_short w, h;
  short x, y;
  u_char nPlanes;
  u_char masking;
  u_char compression;
  u_char pad1;
  u_short transparentColor;
  u_char xAspect, yAspect;
  short pageWidth, pageHeight;
} ILBMbmhd;
enum MaskingType {
  mskNone, mskHasMask, mskHasTransparentColor, mskLasso
};
enum CompressionType {
  cmpNone, cmpByteRun1, cmpRGBN = 4
};

static void
valBMHD(ILBMbmhd * bmhd)
{
  bmhd->w = ntohs(bmhd->w);
  bmhd->h = ntohs(bmhd->h);
  bmhd->x = ntohs(bmhd->x);
  bmhd->y = ntohs(bmhd->y);
  bmhd->transparentColor = ntohs(bmhd->transparentColor);
  bmhd->pageWidth = ntohs(bmhd->pageWidth);
  bmhd->pageHeight = ntohs(bmhd->pageHeight);
}

#if DEBUG
static void
dumpBMHD(const ILBMbmhd * bmhd)
{
  printf("%ubit image %ux%u, max. %lu colors, %s\n",
         bmhd->nPlanes, bmhd->w, bmhd->h,
         1L << bmhd->nPlanes,
         (bmhd->nPlanes == bitppRGB) ? "RGB" : "indexed"
    );
  printf("x/y = %d/%d, Masking:%u, Compr.:%u\n",
         bmhd->x, bmhd->y,
         bmhd->masking, bmhd->compression
    );
  printf("Transparent color: %u, Aspect: %u/%u\n",
         bmhd->transparentColor,
         bmhd->xAspect, bmhd->yAspect
    );
  printf("Page size: %dx%d\n",
         bmhd->pageWidth, bmhd->pageHeight
    );
}
#endif

typedef struct {
  u_long viewModes;
} ILBMcamg;

static void
valCAMG(ILBMcamg * camg)
{
  camg->viewModes = ntohl(camg->viewModes);
}

enum ViewModeFlags {
  lace = 0x00000004,
  extraHalfbrite = 0x00000080,
  ham = 0x00000800,
  hiRes = 0x00008000
};

#if DEBUG
static void
dumpCAMG(const ILBMcamg * camg)
{
  printf("ViewModes: 0x%08lx  ", camg->viewModes);
  if(camg->viewModes & hiRes)
    fputs("HIRES ", stdout);
  if(camg->viewModes & ham)
    fputs("HAM ", stdout);
  if(camg->viewModes & extraHalfbrite)
    fputs("EHB ", stdout);
  if(camg->viewModes & lace)
    fputs("LACE ", stdout);
  fputs("\n",stdout);
}
#endif

typedef struct {
  u_short dpiX;
  u_short dpiY;
} ILBMdpi;

static void
valDPI(ILBMdpi * dpi)
{
  dpi->dpiX = ntohs(dpi->dpiX);
  dpi->dpiY = ntohs(dpi->dpiY);
}

#if DEBUG
static void
dumpDPI(const ILBMdpi * dpi)
{
  printf("Scan resolution: %hux%hu dpi\n",
         dpi->dpiX, dpi->dpiY);
}
#endif

typedef struct {
  ILBMbmhd bmhd;
  ILBMcamg camg;
  u_short cmapSet;
  guchar cmap[maxIndexedColors * byteppRGB];
} ILBMAttribs;


/*** ByteRun1 encoding/decoding ***/


#define MINRUN 3
#define MAXRUN 128
#define MAXDAT 128

static short
unpackRow(FILE * fil, char *bitlinebuf, int bytesNeeded)
{
  short success = 1;
  while (bytesNeeded > 0) {
    int ch;
    ch = fgetc(fil);
    if (EOF == ch) {
      fputs("Got EOF while decrunching!\n", stderr);
      return (0);
    }
    ch = (char) ch;
    if (ch >= 0) {
      ++ch;
      (void) fread(bitlinebuf, ch, 1, fil);
      bitlinebuf += ch;
      bytesNeeded -= ch;
    } else if (ch != -128) {
      int val;
      ch = -ch + 1;
      bytesNeeded -= ch;
      val = fgetc(fil);
      if (EOF == val) {
        fputs("Got EOF while decrunching!\n", stderr);
        return (0);
      } else {
        (void) memset (bitlinebuf, val, ch);
        bitlinebuf += ch;
      }
#if 0
      while (ch--)              /* memset? */
        *bitlinebuf++ = val;
#endif
    } else {
      fputs("unpackRow: Got -128!\n", stderr);
    }
  }
  if (bytesNeeded < 0) {
    fprintf(stderr, "%d == bytesNeeded < 0 (file compression corrupt!)\n",
            bytesNeeded);
    success = 0;
  }
  return (success);
}

static void
unpackBits(const u_char * bitlinebuf,
           guchar * destline, int bitnr, int width)
{
  guchar setmask = 1 << bitnr;
  while (width > 0) {
    int checkmask;
    for (checkmask = 1 << 7; width && checkmask; checkmask >>= 1) {
      if (*bitlinebuf & checkmask)
        *destline |= setmask;
      ++destline;
      --width;
    }
    ++bitlinebuf;
  }
}

static void
writeRGB(const grayval * destline, grayval * dest, int numPixels, int bytepp)
{
  while (numPixels--) {
    *dest = *destline++;
    dest += bytepp;             /*4 */
  }
}

#define ham4bitToGray8(val4) ((((val4) << 4) * (short) 17) / 16)
#define hamXbitToGray8(x,valx) (((((unsigned long)(valx)) << (8-(x))) * (short) ((1<<(x))+1)) / (1<<(x)))

static void
deHam(grayval * dest, const palidx * src, int width, unsigned short depth,
      const grayval * cmap, int alpha)
{
  grayval cr = 0, cg = 0, cb = 0;
  guchar bmask;
  depth -= 2;	/* control bits */
  bmask = (1 << depth) - 1;
  if (width && (*src >> depth)) {
    /* FIXME: only once */
    fputs("Warning: HAM line starts with undefined color. Setting to black.\n", stderr);
  }
  while (width--) {
    guchar idx = *src++;
    switch (idx >> depth) {         /* ham5 only */
      default:
        cr = cmap[((short)idx) * 3 + 0];
        cg = cmap[((short)idx) * 3 + 1];
        cb = cmap[((short)idx) * 3 + 2];
        break;
      case 1:
        cb = hamXbitToGray8(depth, idx & bmask);
        break;
      case 2:
        cr = hamXbitToGray8(depth, idx & bmask);
        break;
      case 3:
        cg = hamXbitToGray8(depth, idx & bmask);
        break;
    }
    *dest++ = cr;
    *dest++ = cg;
    *dest++ = cb;
    /* Leave alpha channel untouched */
    if (alpha)
      ++dest;
  }
}

static void
bitExpandStep(guchar * dest, const guchar * src, int width, int step)
{
  /* Expands alpha masks to 0x00/0xFF style lines */
  guchar bit = 1 << 7;
  while (width--) {
    if (*src & bit)
      *dest = opaque;
    else
      *dest = transparent;
    dest += step;
    bit >>= 1;
    if (!bit) {
      bit = 1 << 7;
      ++src;
    }
  }
}

static void
transGray(grayval * dest, int width, const palidx * transGrayTab)
{
  while (width--) {
    *dest = transGrayTab[*dest];
    ++dest;
  }
}

static short
readPlaneRow(FILE * fil, guchar * bitDest, int bytesInPlaneRow, short compression)
{
  short success = 0;
  switch (compression) {
    case cmpByteRun1:
      success = unpackRow(fil, bitDest, bytesInPlaneRow);
      break;
    case cmpRGBN:
      fputs ("(RGBN compression mode 4 found)\n", stderr);
    default:                   /* fall-thru */
      fputs ("Unknown compression mode---reading raw.\n", stderr);
    case cmpNone:
      success = 1 == fread(bitDest, bytesInPlaneRow, 1, fil);
      if (!success)
        perror("readPlaneRow");
      break;
  }
  if (!success)
    fputs("readPlaneRow: Error.\n", stderr);
  return (success);
}


static void
parseLines(FILE * fil, guchar * dest,
           int width, int hereheight,
           const ILBMbmhd * bmhd, const grayval * cmap,
           u_long viewModes, const grayval * grayTrans)
{
  if (!cmap) {
    /* RGB(A) */
    guchar destline [width];     /* +1? */
    guchar bitlinebuf [BYTEPL (width)];
    while (hereheight--) {
      int rgb;
      for (rgb = 0; rgb < byteppRGB; ++rgb) {
        int bitnr;
        (void) memset(destline, 0, sizeof(destline));
        for (bitnr = 0; bitnr < bitppGray; ++bitnr) {
          readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
          unpackBits(bitlinebuf, destline, bitnr, width);
        }
        writeRGB(destline, dest + rgb, width,
                 byteppRGB + (bmhd->masking != mskNone));
      }
      if (bmhd->masking != mskNone) {
        switch (bmhd->masking) {
          case mskHasMask:
            readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
            bitExpandStep(dest + byteppRGB, bitlinebuf, width, byteppRGBA);
            break;
          case mskHasTransparentColor:
            /* Impossible in non-indexed images */
          default:
            {
              int i;
              for (i = 0; i < width; ++i)
                dest[byteppRGB + i * byteppRGBA] = opaque;
            }
            break;
        }
      }
      dest += width * (byteppRGB + (bmhd->masking != mskNone));
    }
  } else if (!(viewModes & ham)) {
    guchar destline [width];
    guchar bitlinebuf [BYTEPL (width)];
    /* indexed */
    while (hereheight--) {
      int bitnr;
      memset(destline, 0, sizeof(destline));
      /* todo: Schleife entlasten */
      for (bitnr = 0; bitnr < bmhd->nPlanes; ++bitnr) {
        readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
        unpackBits(bitlinebuf, destline, bitnr, width);  /* +7/8? */
      }
      if (grayTrans)
        transGray(destline, width, grayTrans);
      if (bmhd->masking == mskNone) {
        memcpy(dest, destline, width);  /* no alpha */
      } else {
        writeRGB(destline, dest, width, byteppGrayA);  /* make space for alpha */
        switch (bmhd->masking) {
          case mskHasMask:
            readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
            bitExpandStep(dest + byteppGray, bitlinebuf, width, byteppGrayA);
            break;
          case mskHasTransparentColor:
            {
              int i;
              guchar *indx = destline;  /*koennte auch dest lesen */
              guchar *toalpha = dest + byteppGray;
              for (i = width; i > 0; --i) {
                if (*indx++ == bmhd->transparentColor)
                  *toalpha = transparent;
                else
                  *toalpha = opaque;
                toalpha += byteppGrayA;
              }
            }
            break;
          default:
            {
              int i;
              for (i = 0; i < width; ++i)
                dest[byteppGray + i * byteppGrayA] = opaque;
            }
            break;
        }
        dest += width;
      }
      dest += width;
    }
  } else {
    guchar destline [width];
    guchar bitlinebuf [BYTEPL (width)];
    /* HAM */
    while (hereheight--) {
      int bitnr;
      memset(destline, 0, sizeof(destline));
      for (bitnr = 0; bitnr < bmhd->nPlanes; ++bitnr) {
        readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
        unpackBits(bitlinebuf, destline, bitnr, width);
      }
      deHam(dest, destline, width, bmhd->nPlanes, cmap, bmhd->masking != mskNone);
      if (bmhd->masking != mskNone) {
        switch (bmhd->masking) {
          case mskHasMask:
            readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
            bitExpandStep(dest + byteppRGB, bitlinebuf, width, byteppRGBA);
            break;
          case mskHasTransparentColor:
            {
              int i;
              guchar *indx = destline;
              guchar *toalpha = dest + byteppRGB;
              for (i = width; i > 0; --i) {
                if (*indx++ == bmhd->transparentColor)
                  *toalpha = transparent;
                else
                  *toalpha = opaque;
                toalpha += byteppRGBA;
              }
            }
            break;
          default:
            {
              int i;
              for (i = 0; i < width; ++i)
                dest[byteppRGB + i * byteppRGBA] = opaque;
            }
            break;
        }
      }
      dest += width * (byteppRGB + (bmhd->masking != mskNone));
    }
  }
}

static u_long
rlong(FILE * fil)
{
  u_long val;
  if (1 != fread(&val, sizeof(val), 1, fil)) {
    fputs("Error reading hunklength.\n", stderr);
    val = 0;
  } else
    val = ntohl(val);
  return val;
}

static guchar *
allocGrayscale(void)
{
  guchar *gbuf = g_new(guchar, maxGrayshades * byteppRGB);
  if (!gbuf) {
    fputs ("Failed to allocate memory for grayscale map\n", stderr);
  } else {
    guchar *gp;
    int i;
#ifdef VERBOSE
    fputs("Composing grayscale colormap...\n", stdout);
#endif
    for (gp = gbuf, i = 0; i < maxGrayshades; ++i) {
      *gp++ = i;
      *gp++ = i;
      *gp++ = i;
    }
  }
  return (gbuf);
}

static short
isGrayscale(const guchar * cmap, int ncols)
{
  while (ncols--) {
    if (!((*cmap == cmap[1]) && (cmap[1] == cmap[2])))
      return (0);
    cmap += byteppRGB;
  }
  return (1);
}

static guchar *
allocGrayKeep(void)
{
  guchar *gt = g_new(guchar, maxGrayshades);
  if (!gt) {
    fputs ("Out of memory.\n", stderr);
  } else {
    int i;
    for (i = maxGrayshades - 1; i >= 0; --i)
      gt[i] = i;
  }
  return (gt);
}

static guchar *
allocGrayTrans(const guchar * cmap, int ncols)
{
  guchar *gt = g_new(guchar, maxGrayshades);
  if (!gt) {
    fputs ("Out of memory.\n", stderr);
  } else {
    int idx = 0;
    while (ncols--) {
      gt[idx] = *cmap;
      ++idx;
      cmap += byteppRGB;
    }
  }
  return (gt);
}

gint32 loadImage(void *thiz, const char *filename)
{
  gint32 imageID = -1;
  char name[strlen(filename) + 10];

  sprintf(name, "Loading %s:", filename);

  {
    FILE *fil;
#ifdef VERBOSE
    printf("=== Loading %s...\n", filename);
#endif
    fil = fopen(filename, "rb");
    if (!fil)
      fprintf(stderr, "File %s not found.\n", filename);
    else {
      char hdrid[4 + 1];
      hdrid[sizeof(hdrid) - 1] = '\0';
      (void) fread(hdrid, 4, 1, fil);
      if (strcmp(hdrid, "FORM"))
        fprintf(stderr, "No FORM header found.\n");
      else {
        (void) rlong(fil);
        (void) fread(hdrid, 4, 1, fil);
        if (strcmp(hdrid, "ILBM") && strcmp(hdrid, "RGBN"))	/* or RGBN */
          fprintf(stderr, "No ILBM/RGBN header found.\n");
        else {
          long hunksize, ohunksize;
          long width = 0, height = 0;
	  /*
          GDrawable *drawable;
          GPixelRgn pixelRegion;
	  */
          ILBMbmhd bmhd;
          guchar *cmap = 0;
          guchar *grayTrans = 0;
          int ncols = 0;
          ILBMcamg camg =
          {0x00000000};
          bmhd.nPlanes = 0;     /* == not yet used */
          while (1 == fread(hdrid, sizeof(hdrid) - 1, 1, fil)) {
            ohunksize = rlong(fil);
            hunksize = ((ohunksize + 1) / 2) * 2;
#ifdef VERBOSE
            printf("Header %s of %ld byte size.\n", hdrid, ohunksize);
#endif
            if (!strcmp(hdrid, "BMHD")) {
              (void) fread(&bmhd, sizeof(ILBMbmhd), 1, fil);
              valBMHD(&bmhd);
              width = bmhd.w;
              height = bmhd.h;
#ifdef VERBOSE
              dumpBMHD(&bmhd);
#endif
            } else if (!strcmp(hdrid, "CMAP")) {
              ncols = ohunksize / byteppRGB;
#ifdef VERBOSE
              printf("%d colors in CMAP.\n", ncols);
#endif
              cmap = (u_char *) g_new(guchar, ncols * byteppRGB + 1 /*pad */ );
              (void) fread(cmap, hunksize, 1, fil);
            } else if (!strcmp(hdrid, "CAMG")) {
              (void) fread(&camg, sizeof(ILBMcamg), 1, fil);
              valCAMG(&camg);
#ifdef VERBOSE
              dumpCAMG(&camg);
#endif
            } else if (!strcmp(hdrid, "DPI ")) {
              ILBMdpi dpi;
              (void) fread(&dpi, sizeof(ILBMdpi), 1, fil);
              valDPI(&dpi);
#ifdef VERBOSE
              dumpDPI(&dpi);
#endif
            } else if (!strcmp(hdrid, "ANNO") || !strcmp(hdrid, "(C) ")) {
              char cont[ohunksize + 1 + 1];
              (void) fread(cont, hunksize, 1, fil);
              cont[ohunksize] = '\0';
#ifdef VERBOSE
              printf("%s: %s\n", hdrid, cont);
#endif
            } else if (!strcmp(hdrid, "BODY")) {
              guchar *buf;
              short exportRGB, isGray = 0;
              if (!(camg.viewModes & ham) && (6 == bmhd.nPlanes)
                  && (16 == ncols)) {
#ifdef VERBOSE
                printf("Guessing HAM6.\n");
#endif
                camg.viewModes |= ham;
              }
              if ((bitppRGB == bmhd.nPlanes) && cmap) {
                /* Well, I've already seen an 24bit image
                 * with a 32color CMAP. :-/ Ignore it. */
                g_free(cmap);
                cmap = 0;
              }
              if (!cmap && (bitppGray == bmhd.nPlanes)) {
                /* Well, seems to be a grayscale. */
                cmap = allocGrayscale();
                ncols = maxGrayshades;
                isGray = 1;
                grayTrans = allocGrayKeep();  /* 1:1 */
              } else if (cmap && !(camg.viewModes & ham)) {
                isGray = isGrayscale(cmap, ncols);
                if (isGray) {
#ifdef VERBOSE
                  printf("Promoting to grayscale.\n");
#endif
                  grayTrans = allocGrayTrans(cmap, ncols);
                }
              }
#ifdef VERBOSE
              printf("isGray:%d\n", (int) isGray);
#endif
              exportRGB = !cmap || (camg.viewModes & ham);
              if (!exportRGB && !cmap)
                fprintf(stderr, "Colormap (CMAP) missing!\n");

              if (camg.viewModes & extraHalfbrite) {
                int i;
                guchar *nCmap;
#ifdef VERBOSE
                printf("Generating EHB (%d->%d) colormap entries...\n",
                       ncols, 2 * ncols);
#endif
                nCmap = g_new(guchar, byteppRGB * 2 * ncols);
                memcpy(nCmap, cmap, byteppRGB * ncols);
                g_free(cmap);
                cmap = nCmap;
                for (i = byteppRGB * ncols - 1; i >= 0; --i) {
                  cmap[byteppRGB * ncols + i] = cmap[i] >> 1;
                }
                ncols *= 2;
              }
	      
              /* !! Caution: Every msk.. needs alpha, but we don't
                 support everyone yet !! */
	      ilbm_set_size(thiz, width, height, exportRGB, isGray, bmhd.masking != mskNone);
              if (!exportRGB && cmap && !isGray)
                ilbm_set_cmap(thiz, cmap, ncols);
              buf = g_new(guchar, width * ((exportRGB ? byteppRGB : byteppGray)
                             + (bmhd.masking != mskNone)));
              if (buf) {
                int actTop;
                for (actTop = 0; actTop < height; actTop++) {
                  parseLines(fil, buf, width, 1,
                             &bmhd, cmap, camg.viewModes, grayTrans);
		  progress_bar(actTop,height);
		  ilbm_put(thiz, buf, actTop);
                }
                g_free(buf);
              }
              break;            /* Stop parsing IFF */
            } else {
              (void) fseek(fil, hunksize, SEEK_CUR);
            }
          }
          if (cmap) {
            if (grayTrans)
              g_free(grayTrans);
            g_free(cmap);
          }
        }
      }
      (void) fclose(fil);
    }
  }
  return (imageID);
}

