/*
 * GTK See -- a image viewer based on GTK+
 * Copyright (C) 1998 Hotaru Lee <jkhotaru@mail.sti.com.cn> <hotaru@163.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * The GDK library in GTK+ has a well support for xpm images.
 * But GDK xpm support cannot read a header information ONLY.
 * So I wrote this simple function. As you can see, the codes
 * are mostly copy/paste from gdkpixmap.c
 */

/*
 * $Id: im_xpm.c,v 1.11 1999/09/25 20:53:09 malekith Exp malekith $
 * modified by malekith
 */

#include "main-c.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "im_xpm.h"
#include "color-parse.h"

int progress_bar(int, int);

typedef int (*cmp_fn_t) (const void *, const void *);

typedef struct {
	gchar *name;
	guchar r, g, b;
} xpm_color;

void fast_1_cpp(int width, int height, FILE * infile, int num_cols,
		xpm_color * colors, guchar * palbuf, guchar * dbuf);
void fast_2_cpp(int width, int height, FILE * infile, int num_cols,
		xpm_color * colors, guchar * dbuf);

gint xpm_seek_string(FILE * infile, const gchar * str, gint skip_comments);
gint xpm_read_string(FILE * infile, gchar ** buffer, guint * buffer_size);
gint xpm_seek_char(FILE * infile, gchar c);
gchar *xpm_extract_color(gchar * buffer);
gchar *xpm_skip_whitespaces(gchar * buffer);
gchar *xpm_skip_string(gchar * buffer);


gboolean xpm_get_header(gchar * filename, image_info * info)
{
	FILE *infile;
	gchar *buffer = NULL;
	gint width, height, ncols, cpp, n;
	guint buffer_size = 0;

	infile = fopen(filename, "r");
	if (infile == NULL)
		return FALSE;

	if (!xpm_seek_string(infile, "XPM", FALSE)) {
		fclose(infile);
		return FALSE;
	}

	if (!xpm_seek_char(infile, '{')) {
		fclose(infile);
		return FALSE;
	}

	xpm_seek_char(infile, '"');
	fseek(infile, -1, SEEK_CUR);
	xpm_read_string(infile, &buffer, &buffer_size);

	n = sscanf(buffer, "%d %d %d %d", &width, &height, &ncols, &cpp);

	/* we have read all information that we need.
	 * So close the file first.
	 */
	fclose(infile);

	if (n < 4 || cpp >= 32)
		return FALSE;

	info->width = width;
	info->height = height;
	info->bpp = cpp == 1 ? 8 : 24;
	g_free(buffer);

	return TRUE;
}

gchar *xpm_skip_whitespaces(gchar * buffer)
{
	gint32 index = 0;

	while (buffer[index] != 0
	       && (buffer[index] == 0x20 || buffer[index] == 0x09))
		index++;

	return &buffer[index];
}

gchar *xpm_skip_string(gchar * buffer)
{
	gint32 index = 0;

	while (buffer[index] != 0 && buffer[index] != 0x20
	       && buffer[index] != 0x09)
		index++;

	return &buffer[index];
}

gint xpm_seek_string(FILE * infile, const gchar * str, gint skip_comments)
{
	char instr[1024];

	while (!feof(infile)) {
		fscanf(infile, "%1023s", instr);
		if (skip_comments == TRUE && strcmp(instr, "/*") == 0) {
			fscanf(infile, "%1023s", instr);
			while (!feof(infile) && strcmp(instr, "*/") != 0)
				fscanf(infile, "%1023s", instr);
			fscanf(infile, "%1023s", instr);
		}
		if (strcmp(instr, str) == 0)
			return TRUE;
	}

	return FALSE;
}

gint xpm_seek_char(FILE * infile, gchar c)
{
	gint b, oldb;

	while ((b = getc(infile)) != EOF) {
		if (c != b && b == '/') {
			b = getc(infile);
			if (b == EOF)
				return FALSE;
			else if (b == '*') {	/* we have a comment */
				b = -1;
				do {
					oldb = b;
					b = getc(infile);
					if (b == EOF)
						return FALSE;
				}
				while (!(oldb == '*' && b == '/'));
			}
		} else if (c == b)
			return TRUE;
	}
	return FALSE;
}

gint xpm_read_string(FILE * infile, gchar ** buffer, guint * buffer_size)
{
	gint c;
	guint cnt = 0;

	if ((*buffer) == NULL) {
		(*buffer_size) = 10 * sizeof(gchar);
		(*buffer) = g_new(gchar, *buffer_size);
	}

	do
		c = getc(infile);
	while (c != EOF && c != '"');

	if (c != '"')
		return FALSE;

	while ((c = getc(infile)) != EOF) {
		if (cnt == (*buffer_size)) {
			guint new_size = (*buffer_size) * 2;
			if (new_size > (*buffer_size))
				*buffer_size = new_size;
			else
				return FALSE;

			(*buffer) = (gchar *) g_realloc((*buffer), *buffer_size);
		}

		if (c != '"')
			(*buffer)[cnt++] = c;
		else {
			(*buffer)[cnt++] = 0;
			return TRUE;
		}
	}

	return FALSE;
}

/* Xlib crashed ince at a color name lengths around 125 */
#define MAX_COLOR_LEN 120

gchar *xpm_extract_color(gchar * buffer)
{
	gint counter, numnames;
	gchar *ptr = NULL, ch, temp[128];
	gchar color[MAX_COLOR_LEN], *retcol;
	gint space;

	counter = 0;
	while (ptr == NULL) {
		if (buffer[counter] == 'c') {
			ch = buffer[counter + 1];
			if (ch == 0x20 || ch == 0x09)
				ptr = &buffer[counter + 1];
		} else if (buffer[counter] == 0)
			return NULL;

		counter++;
	}

	ptr = xpm_skip_whitespaces(ptr);

	if (ptr[0] == 0)
		return NULL;
	else if (ptr[0] == '#') {
		retcol = g_strdup(ptr);
		return retcol;
	}

	color[0] = 0;
	numnames = 0;

	space = MAX_COLOR_LEN - 1;
	while (space > 0) {
		sscanf(ptr, "%127s", temp);

		if (((gint) ptr[0] == 0) ||
		    (strcmp("s", temp) == 0) || (strcmp("m", temp) == 0) ||
		    (strcmp("g", temp) == 0) || (strcmp("g4", temp) == 0)) {
			break;
		} else {
			if (numnames > 0) {
				space -= 1;
				strcat(color, " ");
			}
			if (space > 0) {
				strncat(color, temp, space);
				space -= MIN(space, (int) strlen(temp));
			}
			ptr = xpm_skip_string(ptr);
			ptr = xpm_skip_whitespaces(ptr);
			numnames++;
		}
	}

	retcol = g_strdup(color);
	return retcol;
}
int xpm_color_compare(const xpm_color * l, const xpm_color * r)
{
	return strcmp(l->name, r->name);
}

gboolean
xpm_load(gchar * filename, int dbuf_len, guchar * dbuf, guchar * palbuf)
{
	FILE *infile = NULL;
	xpm_color *colors, actcolor;
	gint transparent = -1;
	GdkColor gdkcolor;
	gint width, height, num_cols, cpp, cnt, n, ns, xcnt, ycnt;
	gchar *buffer = NULL, pixel_str[32];
	guint buffer_size = 0;
	gulong index;
	gchar *color_name;

	infile = fopen(filename, "r");
	if (infile == NULL)
		return FALSE;

	if (!xpm_seek_string(infile, "XPM", FALSE)) {
		fclose(infile);
		return FALSE;
	}

	if (!xpm_seek_char(infile, '{')) {
		fclose(infile);
		return FALSE;
	}

	xpm_seek_char(infile, '"');
	fseek(infile, -1, SEEK_CUR);
	xpm_read_string(infile, &buffer, &buffer_size);

	n =
	    sscanf(buffer, "%d %d %d %d", &width, &height, &num_cols,
		   &cpp);

	if (n < 4 || cpp >= 32
	    || dbuf_len < width * height * (cpp > 1 ? 3 : 1)) {
		fclose(infile);
		return FALSE;
	}

	g_free(buffer);
	buffer = NULL;

	colors = g_new(xpm_color, num_cols);
	for (cnt = 0; cnt < num_cols; cnt++) {
		xpm_seek_char(infile, '"');
		fseek(infile, -1, SEEK_CUR);
		xpm_read_string(infile, &buffer, &buffer_size);

		colors[cnt].name = g_malloc(sizeof(gchar) * cpp + 1);
		strncpy(colors[cnt].name, buffer, cpp);
		colors[cnt].name[cpp] = '\0';

		color_name = xpm_extract_color(&buffer[cpp]);

		if (color_name != NULL) {
			if (my_color_parse(color_name, &gdkcolor) == FALSE) {
				transparent = cnt;
			} else {
				colors[cnt].r = gdkcolor.red >> 8;
				colors[cnt].g = gdkcolor.green >> 8;
				colors[cnt].b = gdkcolor.blue >> 8;
			}
			g_free(color_name);
		}
		g_free(buffer);
		buffer = NULL;
	}

	index = 0;

	if (cpp == 1) {
		fast_1_cpp(width, height, infile, num_cols, colors, palbuf,
			   dbuf);
		goto finish;
	} else if (cpp == 2) {
		fast_2_cpp(width, height, infile, num_cols, colors, dbuf);
		goto finish;
	}

	qsort(colors, num_cols, sizeof(colors[0]),
	      (cmp_fn_t) xpm_color_compare);
	actcolor.name = pixel_str;
	pixel_str[cpp] = '\0';

	for (ycnt = 0; ycnt < height; ycnt++) {
		xpm_read_string(infile, &buffer, &buffer_size);

		for (n = 0, cnt = 0, xcnt = 0; n < (width * cpp);
		     n += cpp, xcnt++) {
			xpm_color *p;
			strncpy(pixel_str, &buffer[n], cpp);
			ns = 0;
			p = (xpm_color *)
			    bsearch(&actcolor, colors, num_cols, 
					sizeof(xpm_color),
				    	(cmp_fn_t) xpm_color_compare);
			if (p == NULL)	/* screwed up XPM file */
				ns = transparent;
			else
				ns = p - colors;

			*dbuf++ = colors[ns].r;
			*dbuf++ = colors[ns].g;
			*dbuf++ = colors[ns].b;
		}
		g_free(buffer);
		buffer = NULL;
		if (progress_bar(ycnt, height))
			break;
	}

      finish:
	fclose(infile);

	for (n = 0; n < num_cols; n++)
		g_free(colors[n].name);
	g_free(colors);

	return TRUE;
}

void fast_1_cpp(int width, int height, FILE * infile, int num_cols,
		xpm_color * colors, guchar * palbuf,
		guchar * dbuf)
{
	gchar *buffer = NULL;
	guint buffer_size = 0;
	int i;
	if (num_cols > 256)
		num_cols = 256;	// ? ? ?
	for (i = 0; i < num_cols; i++) {
		int c = (guchar) colors[i].name[0];
		palbuf[c * 3 + 0] = colors[i].r;
		palbuf[c * 3 + 1] = colors[i].g;
		palbuf[c * 3 + 2] = colors[i].b;
	}

	for (i = 0; i < height; i++) {
		xpm_read_string(infile, &buffer, &buffer_size);
		memcpy(dbuf, buffer, width);
		g_free(buffer);
		buffer = NULL;
		dbuf += width;
		if (progress_bar(i, height))
			break;
	}
}
void fast_2_cpp(int width, int height, FILE * infile, int num_cols,
		xpm_color * colors, guchar * dbuf)
{
	guchar *lookup[3];
	gchar *buffer = NULL, *p;
	guint buffer_size = 0;
	int i, j;
	// allocate a bit of memory ;)
	lookup[0] = g_malloc(256 * 256);
	lookup[1] = g_malloc(256 * 256);
	lookup[2] = g_malloc(256 * 256);
	for (i = 0; i < num_cols; i++) {
		int c = ((int) ((guchar) colors[i].name[0]) << 8) +
		    (int) ((guchar) colors[i].name[1]);
		lookup[0][c] = colors[i].r;
		lookup[1][c] = colors[i].g;
		lookup[2][c] = colors[i].b;
	}

	for (i = 0; i < height; i++) {
		xpm_read_string(infile, &buffer, &buffer_size);
		p = buffer;
		for (j = 0; j < width; j++) {
			int c = (int) ((guchar) * p++) << 8;
			c += (guchar) * p++;
			*dbuf++ = lookup[0][c];
			*dbuf++ = lookup[1][c];
			*dbuf++ = lookup[2][c];
		}
		g_free(buffer);
		buffer = NULL;
		if (progress_bar(i, height))
			break;
	}
	g_free(lookup[0]);
	g_free(lookup[1]);
	g_free(lookup[2]);
}
